import os
import pandas as pd

from opat.stats import (cum_return,
                        vami,
                        period_return,
                        annualized_return,
                        annualized_std,)

from opat.portfolio import (create_holdings, create_pnl, create_nav)


def read_ts_csv(filepath):
    """
    Load timeseries from csv, the first column must be date indices
    ----------
    filepath: path to the time series csv file

    Returns
    -------
    time indexed Pandas Dataframe
    """
    result = pd.read_csv(filepath,
                         parse_dates=[0],
                         header=0,
                         index_col=0)

    return(result)


__location__ = os.path.realpath(os.path.join(
    os.getcwd(), os.path.dirname(__file__)))

returns_data = read_ts_csv(__location__ + '/test_data/fund_return.csv')
trade_data = pd.read_csv(__location__ + '/test_data/trades.csv',
                         parse_dates=[0],
                         header=0)
price_data = pd.read_csv(__location__ + '/test_data/prices.csv',
                         parse_dates=[0],
                         header=0)
flow_data = pd.read_csv(__location__ + '/test_data/flows.csv',
                        parse_dates=[0],
                        header=0)
print(trade_data.head())
print(returns_data.head())
print(cum_return(returns_data).head())
print(vami(returns_data).head())
print(period_return(returns_data, "month").head())
print(period_return(returns_data, "week").head())
print(period_return(returns_data, "year").head())
print(period_return(returns_data, "quarter").head())
print(annualized_return(returns_data))
print(annualized_std(returns_data))
print(create_holdings(trade_data, price_data).head())
print(create_holdings(trade_data).head())
print(create_pnl(trade_data, price_data).head())
print(create_nav(trade_data, price_data, flow_data))



#
# Copyright 2018 Shawn Lin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
from datetime import datetime, timedelta


# Return related statistics
def total_return(returns):
    """
    Compute total return from simple returns.

    Parameters
    ----------
    returns : pd.Series of periodic returns

    Returns
    -------
    total_returns : array-like
        Series of total returns.
    """
    if len(returns) < 1:
        return returns.copy()

    # Allocate Memory
    result = returns.copy()

    # Compute cumulative return
    result = result.add(1, fill_value=0)
    result = result.prod(skipna=True)
    result = result.add(-1)

    return result


def cum_return(returns):
    """
    Compute cumulative returns from simple returns.

    Parameters
    ----------
    returns : pd.Series of periodic returns

    Returns
    -------
    cumulative_returns : array-like
        Series of cumulative returns.
    """
    if len(returns) < 1:
        return returns.copy()

    # Allocate Memory
    result = returns.copy()

    # Compute cumulative return
    result = result.add(1, fill_value=0)
    result = result.cumprod(skipna=True)
    result = result.add(-1)

    return result


def vami(returns, starting_value=1000):
    """
    Compute VAMI (Value Added Monthly Index) from simple returns.

    Parameters
    ----------
    returns : pd.Series, np.ndarray, or pd.DataFrame
        Returns of the strategy as a percentage, noncumulative.
         - Time series with decimal returns.
         - Example::
            2015-07-16   -0.012143
            2015-07-17    0.045350
            2015-07-20    0.030957
            2015-07-21    0.004902
         - Also accepts two dimensional data. In this case, each column is
           cumulated.
    starting_value: float, optional
       The starting returns.

    Returns
    -------
    vami : array-like
        Series of cumulative returns.
    """
    result = cum_return(returns)
    result = result.add(1)
    result = result.multiply(starting_value)

    return result


def weekly_return(returns):
    """
    Compute weekly returns from higher frequency returns

    Parameters
    ----------
    returns : pd.Series of returns with periodicity shorter than a week

    Returns
    -------
    weekly_returns : array-like
        Series of weekly returns.
    """
    if len(returns) < 1:
        return returns.copy()

    # Allocate Memory
    result = returns.copy()

    # Compute cumulative return
    result = result.add(1, fill_value=0)
    result = result.groupby(pd.Grouper(freq="W")).prod()
    result = result - 1

    return result


def monthly_return(returns):
    """
    Compute monthly returns from higher frequency returns

    Parameters
    ----------
    returns : pd.Series of returns with periodicity shorter than a month

    Returns
    -------
    monthly_returns : array-like
        Series of monthly returns.
    """
    if len(returns) < 1:
        return returns.copy()

    # Allocate Memory
    result = returns.copy()

    # Compute cumulative return
    result = result.add(1, fill_value=0)
    result = result.groupby(pd.Grouper(freq="M")).prod()
    result = result - 1

    return result


def quarterly_return(returns):
    """
    Compute quarterly returns from higher frequency returns

    Parameters
    ----------
    returns : pd.Series of returns with periodicity shorter than a quarter

    Returns
    -------
    quarterly_returns : array-like
        Series of quarterly returns.
    """
    if len(returns) < 1:
        return returns.copy()

    # Allocate Memory
    result = returns.copy()

    # Compute cumulative return
    result = result.add(1, fill_value=0)
    result = result.groupby(pd.Grouper(freq="Q")).prod()
    result = result - 1

    return result


def annual_return(returns):
    """
    Compute annual returns from higher frequency returns

    Parameters
    ----------
    returns : pd.Series of returns with periodicity shorter than a year

    Returns
    -------
    annual_returns : array-like
        Series of annual returns.
    """
    if len(returns) < 1:
        return returns.copy()

    # Allocate Memory
    result = returns.copy()

    # Compute cumulative return
    result = result.add(1, fill_value=0)
    result = result.groupby(pd.Grouper(freq="A")).prod()
    result = result - 1

    return result


def period_return(returns, period):
    """
    Convert higher frequency returns

    Parameters
    ----------
    returns : pd.Series of returns with higher frequency than the target
      periodicity
    period : the target periodicity to convert the returns to, options are
        - week
        - month
        - quarter
        - year

    Returns
    -------
    monthly_returns : array-like
        Series of annual returns.
    """
    return_func = {
        "week": weekly_return,
        "month": monthly_return,
        "quarter": quarterly_return,
        "year": annual_return,
    }

    result = returns.copy()
    result = return_func[period](returns)

    return result


def annualized_return(returns, start_date=None, end_date=None):
    """
    Convert periodic returns into annualized return

    Parameters
    ----------
    returns : pd.Series of returns
    start_date end_date : string in %Y%m%d. Defaults to None. If given, use
    start or end as the start or end date of the series. This is useful for series
    that's already in a lower frequency (e.g. monthly returns) but the exact
    start or end dates are known. Providing start end in this case will generate
    more accurate annualized returns.

    Returns
    -------
    annualized_returns : array-like
        Series of annualized returns.
    """
    result = returns.copy()

    if start_date is None:
        start_date = returns.index[0]
    else:
        start_date = datetime.strptime(start_date, "%Y-%m-%d")
    if end_date is None:
        end_date = returns.index[-1]
    else:
        end_date = datetime.strptime(end_date, "%Y-%m-%d")

    diff_in_years = ((end_date - start_date).total_seconds() / timedelta(days=365.25).total_seconds())

    result = total_return(result)
    result = (result.add(1) ** (1 / diff_in_years)) - 1

    return result


# Risk related statistics
def annualized_std(returns, start_date=None, end_date=None):
    """
    Compute annualized standard deviation (volatility) from periodic returns

    Parameters
    ----------
    returns : pd.Series of returns
    start_date or end_date : string in %Y%m%d. Defaults to None. If given, use
    start or end as the start or end date of the series. This is useful for series
    that's already in a lower frequency (e.g. monthly returns) but the exact
    start or end dates are known. Providing start end in this case will generate
    more accurate annualized standard deviations.

    Returns
    -------
    annualized_returns : array-like
        Series of annualized returns.
    """
    result = returns.copy()

    if start_date is None:
        start_date = returns.index[0]
    else:
        start_date = datetime.strptime(start_date, "%Y-%m-%d")
    if end_date is None:
        end_date = returns.index[-1]
    else:
        end_date = datetime.strptime(end_date, "%Y-%m-%d")

    diff_in_years = ((end_date - start_date).total_seconds() / timedelta(days=365.25).total_seconds())

    result = result.std() * ((result.count() / diff_in_years) ** 0.5)

    return result



#
# Copyright 2019 Shawn Lin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from datetime import datetime


def create_holdings(trades, splits=None):
    """ Aggregate trade data to create day by date holdings information

    Arguments:
        trades {DataFrame} -- trade records
            trades should have the following columns:
            - tradeday: this is the index of the dataframe of trade dates
            - ticker: the name/ticker of the product being traded
            - quantity: number of contracts bought/sold
            - action: whether this is buy/sell



    Keyword Arguments:
        splits {DataFrame} -- split records for each ticker (default: {None})
            splits should have the following columns:
            - tradeday
            - ticker
            - split: the split ratio

    Returns:
        [DataFrame] -- holdings data in the following format:
            - tradeday
            - ticker
            - quantity: number of contracts held
    """

    trades_use = trades.copy()

    # Set start_date and end_date
    start_date = trades["tradeday"].min().date()
    end_date = datetime.now().date()

    # Merge the action and quantity column into 1
    # First combine each day's trading into 1 number by contract,
    # then use the cumulative sum to find out the holdings
    trades_use["quantity"] = trades_use["quantity"] * \
        (trades_use["action"].map({"Buy": 1, "Sell": -1}))
    holdings = trades_use.groupby(["tradeday", "ticker"])["quantity"].sum()
    holdings = holdings.groupby(["ticker"]).cumsum()
    holdings = holdings.reset_index()

    # Expand to daily holdings
    # First expand to wide form and fill each column, then reset index to the
    # desired date series and gathers back to stacked form
    holdings = holdings.pivot(index="tradeday", columns="ticker")
    holdings = holdings.fillna(method="ffill")
    dates = pd.date_range(start_date, end_date, freq='B')
    dates.name = "tradeday"
    holdings = holdings.reindex(dates, method="ffill")
    holdings = holdings.stack("ticker")
    holdings = holdings.reset_index()

    # Keep only non-zero holdings dates, in case contracts are sold
    holdings = holdings[holdings["quantity"] != 0]

    if splits is not None:

        splits_use = splits.copy()

        # Merge with splits and add split to holdings
        splits_use = splits_use[["tradeday", "ticker", "split"]]
        splits_use = holdings.merge(
            splits_use, how="left", on=["tradeday", "ticker"])
        splits_use["split"] = splits_use["split"].fillna(value=1)

        holdings = pd.DataFrame()

        for _, value in splits_use.groupby("ticker"):
            value = value.set_index("tradeday")
            value["prev_holding"] = value["quantity"].shift(1, fill_value=0)
            value["quantity"] = (value["prev_holding"] * \
                (value["split"] - 1)).cumsum() + value["quantity"]
            value["quantity"] = value["quantity"].apply(np.floor)
            value = value.reset_index()
            holdings = holdings.append(
                value[["tradeday", "ticker", "quantity"]], ignore_index=True)

    holdings = holdings.sort_values(
        by=["tradeday", "ticker"]).reset_index(drop=True)
    return holdings


def create_pnl(trades, prices):
    """Create daily portfolio dollar pnl from holdings and trades

    Arguments:
        trades {DataFrame} -- Daily trade data
        prices {DataFrame} -- Daily price data, with dividend and split information
    """

    trades_use = trades.copy()
    prices_use = prices.copy()

    # Create Holdings from trades
    holdings = create_holdings(trades_use, prices_use)

    # Merge holdings and trades with price data
    holdings = holdings.merge(prices_use, how="left",
                              on=["tradeday", "ticker"])
    trades_use = trades_use.merge(
        prices_use, how="left", on=["tradeday", "ticker"])

    # Calculate pnl from holdings and new trades
    holdings_pnl = pd.DataFrame()
    trades_pnl = pd.DataFrame()

    for _, value in holdings.groupby("ticker"):
        value = value.set_index("tradeday")
        value["close"] = value["close"].fillna(method="ffill")
        value["prev_holding"] = value["quantity"].shift(1, fill_value=0)
        value["price_change"] = value["close"] * value["split"] - value["close"].shift(1)
        value["pnl"] = value["price_change"] * value["prev_holding"] + \
            value["dividend"] * value["prev_holding"]
        value = value.reset_index()
        holdings_pnl = holdings_pnl.append(
            value[["tradeday", "ticker", "pnl"]], ignore_index=True)

    for _, value in trades_use.groupby("ticker"):
        value = value.set_index("tradeday")
        value["close"] = value["close"].fillna(method="ffill")
        value["price_change"] = value["close"] - value["price"]
        value["pnl"] = value["price_change"] * value["quantity"] * \
            value["action"].map({"Buy": 1, "Sell": -1})
        value = value.reset_index()
        trades_pnl = trades_pnl.append(
            value[["tradeday", "ticker", "pnl"]], ignore_index=True)

    # Combine pnl into pnl by ticker
    pnl = holdings_pnl.append(trades_pnl, ignore_index=True)
    pnl = pnl.groupby(["tradeday", "ticker"]).sum()

    return pnl


def create_nav(trades, prices, flows):
    """Create dollar nav for each position

    Arguments:
        trades {DataFrame} -- Daily trade data
        prices {DataFrame} -- Daily price data, with dividend and split information
        flows {DataFrame} --  Cash flow data of deposit and withdrawl
    """

    trades_use = trades.copy()
    prices_use = prices.copy()
    flows_use = flows.copy()

    # Create holdings
    holdings = create_holdings(trades_use, prices_use)
    holdings = holdings.merge(prices_use, how="left",
                              on=["tradeday", "ticker"])
    holdings["close"] = holdings.groupby(
        ["ticker"])["close"].fillna(method="ffill")

    # Start date of nav is the first day of flows
    # End date of nav is the last day we have holdings
    start_date = flows_use["tradeday"].min().date()
    end_date = holdings["tradeday"].max().date()

    # Create empty dataframe of dates for merging with
    # nav data later
    dates = pd.date_range(start_date, end_date, freq='B')
    dates.name = "tradeday"

    # Create daily cumulative cashflow resulted from
    # deposit and withdrawal
    cash = flows_use[["tradeday", "amount"]]
    cash = cash.groupby(["tradeday"]).sum()
    cash["nav"] = cash["amount"].cumsum()
    cash = cash["nav"]
    cash = cash.reindex(dates, method="ffill")

    # Create daily cumulative dividend payout information
    dividend = pd.DataFrame()
    for _, value in holdings.groupby("ticker"):
        value = value.set_index("tradeday")
        value["prev_holding"] = value["quantity"].shift(1, fill_value=0)
        value["nav"] = value["dividend"] * value["prev_holding"]
        value = value.reset_index()
        dividend = dividend.append(
            value[["tradeday", "nav"]], ignore_index=True)
    dividend = dividend.groupby(["tradeday"]).sum()
    dividend["nav"] = dividend["nav"].cumsum()
    dividend = dividend.reindex(dates, method="ffill")

    # Create daily cumulative cashflow resulted from trading
    trades_use["nav"] = -trades_use["price"] * trades_use["quantity"] * \
        trades_use["action"].map({"Buy": 1, "Sell": -1})
    trades_use = trades_use.groupby(["tradeday"]).sum()
    trades_use["nav"] = trades_use["nav"].cumsum()
    trades_use = trades_use.reindex(dates, method="ffill")

    # Combine cumulative cashflows from deposit, withdrawl, dividends and
    # trading together into daily cash balances
    cash = cash.add(trades_use["nav"], fill_value=0).add(
        dividend["nav"], fill_value=0)
    cash = cash.reset_index()
    cash["type"] = "cash"
    cash["ticker"] = ""
    cash = cash[["tradeday", "type", "ticker", "nav"]]

    # Create market to market daily holdings' nav
    holdings["nav"] = holdings.groupby(["tradeday", "ticker"], group_keys=False).apply(
        lambda x: x["quantity"] * x["close"])
    holdings["type"] = "equity"
    holdings = holdings[["tradeday", "type", "ticker", "nav"]]

    # Combine cash balances with holding balances
    nav = cash.append(holdings, ignore_index=True).sort_values(
        by=["tradeday", "type", "ticker"]).reset_index(drop=True)

    return nav



import sys

from setuptools import setup

import versioneer

DISTNAME = 'opat'
DESCRIPTION = "opat is a Python library for evaluatoin of financial portfolios"
LONG_DESCRIPTION = """opat is a Python library for evaluatoin of financial
portfolios. opat standands for Open Portfolio Analysis Tool.
"""
MAINTAINER = 'Shawn Lin'
MAINTAINER_EMAIL = 'shawnlin.xl@gmail.com'
AUTHOR = 'Shawn Lin'
AUTHOR_EMAIL = 'shawnlin.xl@gmail.com'
URL = "https://github.com/shawnlinxl/opat"
LICENSE = "Apache License, Version 2.0"
VERSION = "0.0.2"

classifiers = ['Development Status :: 1 - Planning',
               'Programming Language :: Python',
               'Programming Language :: Python :: 3',
               'Programming Language :: Python :: 3.7',
               'License :: OSI Approved :: Apache Software License',
               'Intended Audience :: Science/Research',
               'Topic :: Scientific/Engineering',
               'Topic :: Scientific/Engineering :: Mathematics',
               'Operating System :: OS Independent']

if (sys.version_info.major, sys.version_info.minor) >= (3, 3):
    support_ipython_6 = True
else:
    support_ipython_6 = False

install_reqs = [
    'numpy>=1.11.1',
    'pandas>=0.18.1',
]

test_reqs = []

if __name__ == "__main__":
    setup(
        name=DISTNAME,
        cmdclass=versioneer.get_cmdclass(),
        version=versioneer.get_version(),
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        description=DESCRIPTION,
        license=LICENSE,
        url=URL,
        long_description=LONG_DESCRIPTION,
        packages=['opat', 'opat.tests'],
        classifiers=classifiers,
        install_requires=install_reqs,
        tests_require=test_reqs,
        test_suite='nose.collector',
    )
